/*
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or (at
 *  your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 *  MA 02110-1301, USA.
 *
 *  Copyright © 2006 Pierre Habouzit
 */

/*
 * Copyright notice from original mutt:
 * Copyright (C) 1999-2000 Thomas Roessler <roessler@does-not-exist.org>
 *
 * This file is part of mutt-ng, see http://www.muttng.org/.
 * It's licensed under the GNU General Public License,
 * please see the file GPL in the top level source directory.
 */

/*
 * Yet another MIME encoding for header data.  This time, it's
 * parameters, specified in RFC 2231, and modeled after the
 * encoding used in URLs.
 * 
 * Additionally, continuations and encoding are mixed in an, errrm,
 * interesting manner.
 *
 */

#include <ctype.h>
#include <string.h>
#include <stdlib.h>

#include <lib-lib/mem.h>
#include <lib-lib/str.h>
#include <lib-lib/ascii.h>

#include <lib-mime/mime.h>

#include "charset.h"

typedef struct rfc2231_param {
    struct rfc2231_param *next;

    char *attribute;
    char *value;
    int idx;
    int encoded;
} rfc2231_param;

DO_INIT(rfc2231_param, rfc2231_param);
static inline void rfc2231_param_wipe(rfc2231_param *param)
{
    p_delete(&param->attribute);
    p_delete(&param->value);
}
DO_NEW(rfc2231_param, rfc2231_param);
DO_DELETE(rfc2231_param, rfc2231_param);

/* TODO: MC: replace with a str_unescape */
static void rfc2231_decode_one(char *dst, const char *src)
{
    while (*src) {
        int h1, h2;

        if (*src == '%'
        && (h1 = hexval(src[1])) >= 0 && (h2 = hexval(src[2])) >= 0)
        {
            *dst++ = (h1 << 4) | h2;
            src += 3;
        } else {
            *dst++ = *src++;
        }
    }

    *dst = '\0';
}

/* read the <charset>'foo' part into charset, and skip that */
static char *rfc2231_get_charset(char *value, char *charset, size_t chslen)
{
    char *t, *u;

    t = strchr(value, '\'');
    if (!t) {
        charset[0] = '\0';
        return value;
    }

    *t = '\0';
    m_strcpy(charset, chslen, value);

    if ((u = strchr(t + 1, '\''))) {
        return u + 1;
    } else {
        return t + 1;
    }
}


/* insert parameter into an ordered list.
 * 
 * Primary sorting key: attribute
 * Secondary sorting key: idx
 *
 * XXX: MC: looks very unclear to me
 */
static void
rfc2231_list_insert(rfc2231_param **list, rfc2231_param *par)
{
    int c;

    while (*list) {
        rfc2231_param *q = *list;

        list = &(*list)->next;

        c = m_strcmp(par->value, q->value);
        if ((c > 0) || (c == 0 && par->idx >= q->idx))
            break;
    }

    par->next = *list;
    *list = par;
}

static void purge_empty_parameters(PARAMETER **headp)
{
    while (*headp) {
        PARAMETER *p = *headp;

        if (!p->attribute || !p->value) {
            *headp = p->next;
            p->next = NULL;
            mutt_free_parameter(&p);
        } else {
            headp = &(*headp)->next;
        }
    }
}

/* process continuation parameters */
/* XXX: MC: not read */
static void
rfc2231_join_continuations(PARAMETER **head, rfc2231_param *par)
{
    rfc2231_param *q;

    char attribute[STRING];
    char charset[STRING];
    char *value = NULL;
    char *valp;
    int encoded;

    size_t l, vl;

    while (par) {
        value = NULL;
        l = 0;

        m_strcpy(attribute, sizeof(attribute), par->attribute);

        if ((encoded = par->encoded))
            valp = rfc2231_get_charset (par->value, charset, sizeof (charset));
        else
            valp = par->value;

        do {
            if (encoded && par->encoded)
                rfc2231_decode_one (par->value, valp);

            vl = m_strlen(par->value);

            p_realloc(&value, l + vl + 1);
            strcpy (value + l, par->value);   /* __STRCPY_CHECKED__ */
            l += vl;

            q = par->next;
            rfc2231_param_delete (&par);
            if ((par = q))
                valp = par->value;
        } while (par && !m_strcmp(par->attribute, attribute));

        if (value) {
            if (encoded)
                mutt_convert_string (&value, charset, Charset, M_ICONV_HOOK_FROM);
            *head = mutt_new_parameter ();
            (*head)->attribute = m_strdup(attribute);
            (*head)->value = value;
            head = &(*head)->next;
        }
    }
}

/****************************************************************************/
/* Public API                                                               */
/****************************************************************************/

/* XXX: MC: not read */
void rfc2231_decode_parameters (PARAMETER ** headp)
{
    PARAMETER *head = NULL;
    PARAMETER **last;
    PARAMETER *p, *q;

    rfc2231_param *conthead = NULL;
    rfc2231_param *conttmp;

    char *s, *t;
    char charset[STRING];

    int encoded;
    int idx;
    short dirty = 0;              /* set to 1 when we may have created
                                   * empty parameters.
                                   */

    if (!headp)
        return;

    purge_empty_parameters (headp);

    for (last = &head, p = *headp; p; p = q) {
        q = p->next;

        if (!(s = strchr (p->attribute, '*'))) {

            /* 
             * Using RFC 2047 encoding in MIME parameters is explicitly
             * forbidden by that document.  Nevertheless, it's being
             * generated by some software, including certain Lotus Notes to 
             * Internet Gateways.  So we actually decode it.
             */

            if (option (OPTRFC2047PARAMS) && p->value && strstr (p->value, "=?"))
                rfc2047_decode (&p->value);
            else if (!option (OPTSTRICTMIME)) {
                if (ascii_strcasecmp (AssumedCharset, "us-ascii"))
                    mutt_convert_nonmime_string (&p->value);
            }

            *last = p;
            last = &p->next;
            p->next = NULL;
        }
        else if (*(s + 1) == '\0') {
            *s = '\0';

            s = rfc2231_get_charset (p->value, charset, sizeof (charset));
            rfc2231_decode_one (p->value, s);
            mutt_convert_string (&p->value, charset, Charset, M_ICONV_HOOK_FROM);

            *last = p;
            last = &p->next;
            p->next = NULL;

            dirty = 1;
        }
        else {
            *s = '\0';
            s++;                      /* let s point to the first character of idx. */
            for (t = s; *t && isdigit ((unsigned char) *t); t++);
            encoded = (*t == '*');
            *t = '\0';

            idx = atoi (s);

            conttmp = rfc2231_param_new ();
            conttmp->attribute = p->attribute;
            conttmp->value = p->value;
            conttmp->encoded = encoded;
            conttmp->idx = idx;

            p->attribute = NULL;
            p->value = NULL;
            p_delete(&p);

            rfc2231_list_insert (&conthead, conttmp);
        }
    }

    if (conthead) {
        rfc2231_join_continuations (last, conthead);
        dirty = 1;
    }

    *headp = head;

    if (dirty)
        purge_empty_parameters (headp);
}

#define RFC2231_SPECIALS  "@.,;:<>[]\\\"()?/= \t*'%"

int rfc2231_encode_string(char **s)
{
    char *charset = NULL;
    char *e, *p, *t, *d = NULL;
    int escapes = 0;
    ssize_t dlen = 0;

    /*
     * A shortcut to detect pure 7bit data.
     *
     * This should prevent the worst when character set handling is flawed.
     */

    for (p = *s; ; p++) {
        if (*p & 0x80)
            break;
        if (!*p)
            return 0;
    }

    if (Charset && SendCharset) {
        charset = mutt_choose_charset(Charset, SendCharset,
                                      *s, m_strlen(*s), &d, &dlen);
    }

    if (!charset) {
        charset = m_strdup(Charset ? Charset : "unknown-8bit");
        d = *s;
        dlen = m_strlen(d);
    }

    for (p = d; *p; p++) {
        if (*p < 0x20 || *p >= 0x7f || strchr(RFC2231_SPECIALS, *p)) {
            ++escapes;
        }
    }

    e = p_new(char, dlen + 2 * escapes + m_strlen(charset) + 3);

    t = e + sprintf(e, "%s''", charset);
    for (p = d; *p; p++) {
        if (*p < 0x20 || *p >= 0x7f || strchr(RFC2231_SPECIALS, *p)) {
            *t++ = '%';
            *t++ = __m_b36chars_upper[*p >> 4];
            *t++ = __m_b36chars_upper[*p & 0xf];
        } else {
            *t++ = *p;
        }
    }
    *t = '\0';

    if (d != *s)
        p_delete(&d);
    p_delete(s);
    p_delete(&charset);

    *s = e;
    return 1;
}
